#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Async iterator extensions for paginated data."""
from __future__ import annotations

import abc
import datetime
import typing

from hikari import bases
from hikari.internal import assertions
from hikari.internal import more_collections

_T = typing.TypeVar("_T")


class PaginatedResults(typing.Generic[_T], abc.ABC):
    """A set of results that are fetched asynchronously from the API as needed.

    This is a `typing.AsyncIterable` and `typing.AsyncIterator` with several
    additional helpful methods provided for convenience.

    Examples
    --------
    You can use this in multiple ways.

    As an async iterable:

        >>> async for item in paginated_results:
        ...    process(item)

    As an eagerly retrieved set of results (performs all API calls at once,
    which may be slow for large sets of data):

        >>> results = await paginated_results
        >>> # ... which is equivalent to this...
        >>> results = [item async for item in paginated_results]

    As an async iterator (not recommended):

        >>> try:
        ...    while True:
        ...        process(await paginated_results.__anext__())
        ... except StopAsyncIteration:
        ...    pass

    Additionally, you can make use of some of the provided helper methods
    on this class to perform basic operations easily.

    Iterating across the items with indexes (like `enumerate` for normal
    iterables):

        >>> async for i, item in paginated_results.enumerate():
        ...    print(i, item)
        (0, foo)
        (1, bar)
        (2, baz)

    Limiting the number of results you iterate across:

        >>> async for item in paginated_results.limit(3):
        ...    process(item)
    """

    __slots__ = ()

    def enumerate(self, *, start: int = 0) -> PaginatedResults[typing.Tuple[int, _T]]:
        """Enumerate the paginated results lazily.

        This behaves as an asyncio-friendly version of `builtins.enumerate`
        which uses much less memory than collecting all the results first and
        calling `enumerate` across them.

        Parameters
        ----------
        start : int
            Optional int to start at. If omitted, this is `0`.

        Examples
        --------
            >>> async for i, item in paginated_results.enumerate():
            ...    print(i, item)
            (0, foo)
            (1, bar)
            (2, baz)
            (3, bork)
            (4, qux)

            >>> async for i, item in paginated_results.enumerate(start=9):
            ...    print(i, item)
            (9, foo)
            (10, bar)
            (11, baz)
            (12, bork)
            (13, qux)

            >>> async for i, item in paginated_results.enumerate(start=9).limit(3):
            ...    print(i, item)
            (9, foo)
            (10, bar)
            (11, baz)

        Returns
        -------
        PaginatedResults[typing.Tuple[int, T]]
            A paginated results view that asynchronously yields an increasing
            counter in a tuple with each result, lazily.
        """
        return _EnumeratedPaginatedResults(self, start=start)

    def limit(self, limit: int) -> PaginatedResults[_T]:
        """Limit the number of items you receive from this async iterator.

        Parameters
        ----------
        limit : int
            The number of items to get. This must be greater than zero.

        Examples
        --------
            >>> async for item in paginated_results.limit(3):
            ...     print(item)


        Returns
        -------
        PaginatedResults[T]
            A paginated results view that asynchronously yields a maximum
            of the given number of items before completing.
        """
        return _LimitedPaginatedResults(self, limit)

    def _complete(self) -> typing.NoReturn:
        raise StopAsyncIteration("No more items exist in this paginator. It has been exhausted.") from None

    def __aiter__(self) -> PaginatedResults[_T]:
        # We are our own iterator.
        return self

    async def _fetch_all(self) -> typing.Sequence[_T]:
        return [item async for item in self]

    @staticmethod
    def _prepare_first_id(value, if_none=bases.Snowflake.min()) -> str:
        """Prepare the given first ID type passed by the user.

        Given an object with an ID, a datetime, an int, a snowflake, or a string
        type, convert the element to the string ID snowflake it represents
        that can be passed to the underlying REST API safely.

        Parameters
        ----------
        value
            The element to prepare.
        if_none
            The value to use if the `value` is `None`. Defaults to a snowflake
            of `0`.

        Returns
        -------
        str
            The string ID.
        """
        if value is None:
            value = if_none

        if isinstance(value, datetime.datetime):
            value = str(int(bases.Snowflake.from_datetime(value)))

        if isinstance(value, (int, bases.Snowflake)):
            return str(value)
        if isinstance(value, bases.Unique):
            return str(value.id)
        if isinstance(value, str):
            return value

        raise TypeError("expected object with ID, datetime, snowflake, or None")

    def __await__(self):
        return self._fetch_all().__await__()

    @abc.abstractmethod
    async def __anext__(self) -> _T:
        ...


_EnumeratedT = typing.Tuple[int, _T]


class _EnumeratedPaginatedResults(typing.Generic[_T], PaginatedResults[_EnumeratedT]):
    __slots__ = ("_i", "_paginator")

    def __init__(self, paginator: PaginatedResults[_T], *, start: int) -> None:
        self._i = start
        self._paginator = paginator

    async def __anext__(self) -> typing.Tuple[int, _T]:
        pair = self._i, await self._paginator.__anext__()
        self._i += 1
        return pair


class _LimitedPaginatedResults(typing.Generic[_T], PaginatedResults[_T]):
    __slots__ = ("_paginator", "_count", "_limit")

    def __init__(self, paginator: PaginatedResults[_T], limit: int) -> None:
        assertions.assert_that(limit > 0, "limit must be positive and non-zero")
        self._paginator = paginator
        self._count = 0
        self._limit = limit

    async def __anext__(self) -> _T:
        if self._count >= self._limit:
            self._complete()

        next_item = await self._paginator.__anext__()
        self._count += 1
        return next_item


class BufferedPaginatedResults(typing.Generic[_T], PaginatedResults[_T]):
    """A buffered paginator implementation that handles chunked responses."""

    __slots__ = ("_buffer",)

    def __init__(self) -> None:
        # Start with an empty generator to force the paginator to get the next item.
        self._buffer = more_collections.EMPTY_GENERATOR_EXPRESSION

    @abc.abstractmethod
    async def _next_chunk(self) -> typing.Optional[typing.Generator[typing.Any, None, _T]]:
        # Return `None` when exhausted.
        ...

    async def __anext__(self) -> _T:
        # This sneaky snippet of code lets us use generators rather than lists.
        # This is important, as we can use this to make generators that
        # deserialize loads of items lazy. If we only want 10 messages of
        # history, we can use the same code and prefetch 100 without any
        # performance hit from it other than the JSON string response.
        try:
            return next(self._buffer)
        except StopIteration:
            self._buffer = await self._next_chunk()
            if self._buffer is None:
                self._complete()
            else:
                return next(self._buffer)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Provides a facade around `hikari.net.shards.Shard`.

This handles parsing and initializing the object from a configuration, as
well as restarting it if it disconnects.

Additional functions and coroutines are provided to update the presence on the
shard using models defined in `hikari`.
"""

from __future__ import annotations

__all__ = ["ShardClient", "ShardClientImpl"]

import abc
import asyncio
import logging
import time
import typing

import aiohttp

from hikari import errors
from hikari.clients import runnable
from hikari.clients import shard_states
from hikari.events import other
from hikari.internal import assertions
from hikari.net import codes
from hikari.net import ratelimits
from hikari.net import shards

if typing.TYPE_CHECKING:
    import datetime

    from hikari import gateway_entities
    from hikari import guilds
    from hikari import intents as _intents
    from hikari.clients import components as _components


class ShardClient(runnable.RunnableClient, abc.ABC):
    """Definition of the interface for a conforming shard client."""

    __slots__ = ()

    @property
    @abc.abstractmethod
    def shard_id(self) -> int:
        """Shard ID (this is 0-indexed)."""

    @property
    @abc.abstractmethod
    def shard_count(self) -> int:
        """Count of how many shards make up this bot."""

    @property
    @abc.abstractmethod
    def status(self) -> guilds.PresenceStatus:
        """User status for this shard."""

    @property
    @abc.abstractmethod
    def activity(self) -> typing.Optional[gateway_entities.Activity]:
        """Activity for the user status for this shard.

        This will be `None` if there is no activity.
        """

    @property
    @abc.abstractmethod
    def idle_since(self) -> typing.Optional[datetime.datetime]:
        """Timestamp of when the user of this shard appeared to be idle.

        This will be `None` if not applicable.
        """

    @property
    @abc.abstractmethod
    def is_afk(self) -> bool:
        """Whether the user is appearing as AFK or not.."""

    @property
    @abc.abstractmethod
    def heartbeat_latency(self) -> float:
        """Latency between sending a HEARTBEAT and receiving an ACK in seconds.

        This will be `float("nan")` until the first heartbeat is performed.
        """

    @property
    @abc.abstractmethod
    def heartbeat_interval(self) -> float:
        """Time period to wait between sending HEARTBEAT payloads in seconds.

        This will be `float("nan")` until the connection has received a `HELLO`
        payload.
        """

    @property
    @abc.abstractmethod
    def disconnect_count(self) -> int:
        """Count of number of times this shard's connection has disconnected."""

    @property
    @abc.abstractmethod
    def reconnect_count(self) -> int:
        """Count of number of times this shard's connection has reconnected.

        This includes RESUME and re-IDENTIFY events.
        """

    @property
    @abc.abstractmethod
    def connection_state(self) -> shard_states.ShardState:
        """State of this shard's connection."""

    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Whether the shard is connected or not."""

    @property
    @abc.abstractmethod
    def seq(self) -> typing.Optional[int]:
        """Sequence ID of the shard.

        This is the number of payloads that have been received since the last
        `IDENTIFY` was sent.
        """

    @property
    @abc.abstractmethod
    def session_id(self) -> typing.Optional[str]:
        """Session ID of the shard connection.

        Will be `None` if there is no session.
        """

    @property
    @abc.abstractmethod
    def version(self) -> float:
        """Version being used for the gateway API."""

    @property
    @abc.abstractmethod
    def intents(self) -> typing.Optional[_intents.Intent]:
        """Intents that are in use for the shard connection.

        If intents are not being used at all, then this will be `None` instead.
        """

    @abc.abstractmethod
    async def update_presence(
        self,
        *,
        status: guilds.PresenceStatus = ...,
        activity: typing.Optional[gateway_entities.Activity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        """Update the presence of the user for the shard.

        This will only update arguments that you explicitly specify a value for.
        Any arguments that you do not explicitly provide some value for will
        not be changed.

        !!! warning
            This will fail if the shard is not online.

        Parameters
        ----------
        status : hikari.guilds.PresenceStatus
            If specified, the new status to set.
        activity : hikari.gateway_entities.Activity, optional
            If specified, the new activity to set.
        idle_since : datetime.datetime, optional
            If specified, the time to show up as being idle since, or
            `None` if not applicable.
        is_afk : bool
            If specified, whether the user should be marked as AFK.
        """


class ShardClientImpl(ShardClient):
    """The primary interface for a single shard connection.

    This contains several abstractions to enable usage of the low
    level gateway network interface with the higher level constructs
    in `hikari`.

    Parameters
    ----------
    shard_id : int
        The ID of this specific shard.
    shard_id : int
        The number of shards that make up this distributed application.
    components : hikari.clients.components.Components
        The client components that this shard client should be bound by.
        Includes the the gateway configuration to use to initialize this shard
        and the consumer of a raw event.
    url : str
        The URL to connect the gateway to.

    !!! note
        Generally, you want to use
        `hikari.clients.bot_base.BotBase` rather than this class
        directly, as that will handle sharding where enabled and applicable,
        and provides a few more bits and pieces that may be useful such as state
        management and event dispatcher integration. and If you want to customize
        this, you can subclass it and simply override anything you want.
    """

    __slots__ = (
        "_activity",
        "_components",
        "_connection",
        "_idle_since",
        "_is_afk",
        "_raw_event_consumer",
        "_shard_state",
        "_status",
        "_task",
        "logger",
    )

    def __init__(self, shard_id: int, shard_count: int, components: _components.Components, url: str,) -> None:
        super().__init__(logging.getLogger(f"hikari.{type(self).__qualname__}.{shard_id}"))
        self._components = components
        self._raw_event_consumer = components.event_manager
        self._activity = components.config.initial_activity
        self._idle_since = components.config.initial_idle_since
        self._is_afk = components.config.initial_is_afk
        self._status = components.config.initial_status
        self._shard_state = shard_states.ShardState.NOT_RUNNING
        self._task = None
        self._connection = shards.Shard(
            compression=components.config.gateway_use_compression,
            connector=components.config.tcp_connector,
            debug=components.config.debug,
            dispatch=lambda c, n, pl: components.event_manager.process_raw_event(self, n, pl),
            initial_presence=self._create_presence_pl(
                status=components.config.initial_status,
                activity=components.config.initial_activity,
                idle_since=components.config.initial_idle_since,
                is_afk=components.config.initial_is_afk,
            ),
            intents=components.config.intents,
            large_threshold=components.config.large_threshold,
            proxy_auth=components.config.proxy_auth,
            proxy_headers=components.config.proxy_headers,
            proxy_url=components.config.proxy_url,
            session_id=None,
            seq=None,
            shard_id=shard_id,
            shard_count=shard_count,
            ssl_context=components.config.ssl_context,
            token=components.config.token,
            url=url,
            verify_ssl=components.config.verify_ssl,
            version=components.config.gateway_version,
        )

    @property
    def shard_id(self) -> int:
        return self._connection.shard_id

    @property
    def shard_count(self) -> int:
        return self._connection.shard_count

    @property
    def status(self) -> guilds.PresenceStatus:
        return self._status

    @property
    def activity(self) -> typing.Optional[gateway_entities.Activity]:
        return self._activity

    @property
    def idle_since(self) -> typing.Optional[datetime.datetime]:
        return self._idle_since

    @property
    def is_afk(self) -> bool:
        return self._is_afk

    @property
    def heartbeat_latency(self) -> float:
        return self._connection.heartbeat_latency

    @property
    def heartbeat_interval(self) -> float:
        return self._connection.heartbeat_interval

    @property
    def disconnect_count(self) -> int:
        return self._connection.disconnect_count

    @property
    def reconnect_count(self) -> int:
        return self._connection.reconnect_count

    @property
    def connection_state(self) -> shard_states.ShardState:
        return self._shard_state

    @property
    def is_connected(self) -> bool:
        return self._connection.is_connected

    @property
    def seq(self) -> typing.Optional[int]:
        return self._connection.seq

    @property
    def session_id(self) -> typing.Optional[str]:
        return self._connection.session_id

    @property
    def version(self) -> float:
        return self._connection.version

    @property
    def intents(self) -> typing.Optional[_intents.Intent]:
        return self._connection.intents

    async def start(self):
        """Connect to the gateway on this shard and keep the connection alive.

        This will wait for the shard to dispatch a `READY` event, and
        then return.
        """
        if self._shard_state not in (shard_states.ShardState.NOT_RUNNING, shard_states.ShardState.STOPPED):
            raise RuntimeError("Cannot start a shard twice")

        self._task = asyncio.create_task(self._keep_alive(), name="ShardClient#keep_alive")

        completed, _ = await asyncio.wait(
            [self._task, self._connection.ready_event.wait()], return_when=asyncio.FIRST_COMPLETED
        )

        for task in completed:
            if ex := task.exception():
                raise ex

    async def join(self) -> None:
        """Wait for the shard to shut down fully."""
        if self._task:
            await self._task

    async def close(self) -> None:
        """Request that the shard shuts down.

        This will wait for the client to shut down before returning.
        """
        if self._shard_state != shard_states.ShardState.STOPPING:
            self._shard_state = shard_states.ShardState.STOPPING
            self.logger.debug("stopping shard")

            await self._components.event_dispatcher.dispatch_event(other.StoppingEvent())

            await self._connection.close()

            if self._task is not None:
                await self._task

            await self._components.event_dispatcher.dispatch_event(other.StoppedEvent())

    async def _keep_alive(self):  # pylint: disable=too-many-branches
        back_off = ratelimits.ExponentialBackOff(base=1.85, maximum=600, initial_increment=2)
        last_start = time.perf_counter()
        do_not_back_off = True

        await self._components.event_dispatcher.dispatch_event(other.StartingEvent())

        while True:
            try:
                if not do_not_back_off and time.perf_counter() - last_start < 30:
                    next_back_off = next(back_off)
                    self.logger.info(
                        "restarted within 30 seconds, will backoff for %.2fs", next_back_off,
                    )
                    await asyncio.sleep(next_back_off)
                else:
                    back_off.reset()

                last_start = time.perf_counter()
                do_not_back_off = False

                connect_task = await self._spin_up()

                if self._components.event_dispatcher is not None and self.reconnect_count == 0:
                    # Only dispatch this on initial connect, not on reconnect.
                    await self._components.event_dispatcher.dispatch_event(other.StartedEvent())

                await connect_task
                self.logger.critical("shut down silently! this shouldn't happen!")

            except aiohttp.ClientConnectorError as ex:
                self.logger.exception(
                    "failed to connect to Discord to initialize a websocket connection", exc_info=ex,
                )

            except errors.GatewayZombiedError:
                self.logger.warning("entered a zombie state and will be restarted")

            except errors.GatewayInvalidSessionError as ex:
                if ex.can_resume:
                    self.logger.warning("invalid session, so will attempt to resume")
                else:
                    self.logger.warning("invalid session, so will attempt to reconnect")
                    self._connection.seq = None
                    self._connection.session_id = None

                do_not_back_off = True
                await asyncio.sleep(5)

            except errors.GatewayMustReconnectError:
                self.logger.warning("instructed by Discord to reconnect")
                do_not_back_off = True
                await asyncio.sleep(5)

            except errors.GatewayServerClosedConnectionError as ex:
                if ex.close_code in (
                    codes.GatewayCloseCode.NOT_AUTHENTICATED,
                    codes.GatewayCloseCode.AUTHENTICATION_FAILED,
                    codes.GatewayCloseCode.ALREADY_AUTHENTICATED,
                    codes.GatewayCloseCode.SHARDING_REQUIRED,
                    codes.GatewayCloseCode.INVALID_VERSION,
                    codes.GatewayCloseCode.INVALID_INTENT,
                    codes.GatewayCloseCode.DISALLOWED_INTENT,
                ):
                    self.logger.error("disconnected by Discord, %s: %s", type(ex).__name__, ex.reason)
                    raise ex from None

                self.logger.warning("disconnected by Discord, will attempt to reconnect")

            except errors.GatewayClientDisconnectedError:
                self.logger.warning("unexpected connection close, will attempt to reconnect")

            except errors.GatewayClientClosedError:
                self.logger.warning("shutting down")
                return
            except Exception as ex:
                self.logger.debug("propagating unexpected exception", exc_info=ex)
                raise ex

    async def _spin_up(self) -> asyncio.Task:
        self.logger.debug("initializing shard")
        self._shard_state = shard_states.ShardState.CONNECTING

        is_resume = self._connection.seq is not None and self._connection.session_id is not None

        connect_task = asyncio.create_task(self._connection.connect(), name="Shard#connect")

        completed, _ = await asyncio.wait(
            [connect_task, self._connection.hello_event.wait()], return_when=asyncio.FIRST_COMPLETED
        )

        for task in completed:
            if ex := task.exception():
                raise ex

        self.logger.info("received HELLO, interval is %ss", self._connection.heartbeat_interval)

        completed, _ = await asyncio.wait(
            [connect_task, self._connection.handshake_event.wait()], return_when=asyncio.FIRST_COMPLETED
        )

        for task in completed:
            if ex := task.exception():
                raise ex

        if is_resume:
            self.logger.info("sent RESUME, waiting for RESUMED event")
            self._shard_state = shard_states.ShardState.RESUMING

            completed, _ = await asyncio.wait(
                [connect_task, self._connection.resumed_event.wait()], return_when=asyncio.FIRST_COMPLETED
            )

            for task in completed:
                if ex := task.exception():
                    raise ex

            self.logger.info("now RESUMED")

        else:
            self.logger.info("sent IDENTIFY, waiting for READY event")

            self._shard_state = shard_states.ShardState.WAITING_FOR_READY

            completed, _ = await asyncio.wait(
                [connect_task, self._connection.ready_event.wait()], return_when=asyncio.FIRST_COMPLETED
            )

            for task in completed:
                if ex := task.exception():
                    raise ex

            self.logger.info("now READY")

        self._shard_state = shard_states.ShardState.READY

        return connect_task

    async def update_presence(
        self,
        *,
        status: guilds.PresenceStatus = ...,
        activity: typing.Optional[gateway_entities.Activity] = ...,
        idle_since: typing.Optional[datetime.datetime] = ...,
        is_afk: bool = ...,
    ) -> None:
        # We wouldn't ever want to do this, so throw an error if it happens.
        assertions.assert_that(
            status is not ... or activity is not ... or idle_since is not ... or is_afk is not ...,
            "update_presence requires at least one argument to be passed",
        )

        status = self._status if status is ... else status
        activity = self._activity if activity is ... else activity
        idle_since = self._idle_since if idle_since is ... else idle_since
        is_afk = self._is_afk if is_afk is ... else is_afk

        presence = self._create_presence_pl(status=status, activity=activity, idle_since=idle_since, is_afk=is_afk)
        await self._connection.update_presence(presence)

        # If we get this far, the update succeeded probably, or the gateway just died. Whatever.
        self._status = status
        self._activity = activity
        self._idle_since = idle_since
        self._is_afk = is_afk

    @staticmethod
    def _create_presence_pl(
        status: guilds.PresenceStatus,
        activity: typing.Optional[gateway_entities.Activity],
        idle_since: typing.Optional[datetime.datetime],
        is_afk: bool,
    ) -> typing.Dict[str, typing.Any]:
        return {
            "status": status,
            "idle_since": idle_since.timestamp() * 1000 if idle_since is not None else None,
            "game": activity.serialize() if activity is not None else None,
            "afk": is_afk,
        }

    def __str__(self) -> str:
        return f"Shard {self.shard_id} in pool of {self.shard_count} shards"

    def __repr__(self) -> str:
        return (
            "ShardClient("
            + ", ".join(
                f"{k}={getattr(self, k)!r}"
                for k in ("shard_id", "shard_count", "connection_state", "heartbeat_interval", "heartbeat_latency")
            )
            + ")"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Provides the valid routes that can be used on the API, as well as mechanisms to aid with rate limit bucketing."""

from __future__ import annotations

__all__ = ["CompiledRoute", "RouteTemplate"]

import re
import typing

DEFAULT_MAJOR_PARAMS: typing.Final[typing.Set[str]] = {"channel_id", "guild_id", "webhook_id"}
HASH_SEPARATOR: typing.Final[str] = ";"


class CompiledRoute:
    """A compiled representation of a route ready to be made into a full e and to be used for a request.

    Parameters
    ----------
    route_template : RouteTemplate
        The route template used to make this route.
    path : str
        The path with any major parameters interpolated in.
    major_params_hash : str
        The part of the hash identifier to use for the compiled set of major parameters.
    """

    __slots__ = ("route_template", "major_param_hash", "compiled_path", "hash_code")

    route_template: typing.Final[RouteTemplate]
    """The route template this compiled route was created from."""

    major_param_hash: typing.Final[str]
    """The major parameters in a bucket hash-compatible representation."""

    compiled_path: typing.Final[str]
    """The compiled route path to use."""

    hash_code: typing.Final[int]
    """The hash code."""

    def __init__(self, route_template: RouteTemplate, path: str, major_params_hash: str) -> None:
        self.route_template = route_template
        self.major_param_hash = major_params_hash
        self.compiled_path = path
        self.hash_code = hash((self.method, self.route_template.path_template, major_params_hash))

    @property
    def method(self) -> str:
        """Return the HTTP method of this compiled route."""
        return self.route_template.method

    def create_url(self, base_url: str) -> str:
        """Create the full URL with which you can make a request.

        Parameters
        ----------
        base_url : str
            The base of the URL to prepend to the compiled path.

        Returns
        -------
        str
            The full URL for the route.
        """
        return base_url + self.compiled_path

    def create_real_bucket_hash(self, initial_bucket_hash: str) -> str:
        """Create a full bucket hash from a given initial hash.

        The result of this hash will be decided by the value of the major
        parameters passed to the route during the compilation phase.

        Parameters
        ----------
        initial_bucket_hash : str
            The initial bucket hash provided by Discord in the HTTP headers
            for a given response.

        Returns
        -------
        str
            The input hash amalgamated with a hash code produced by the
            major parameters in this compiled route instance.
        """
        return initial_bucket_hash + HASH_SEPARATOR + self.major_param_hash

    def __hash__(self) -> int:
        return self.hash_code

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, CompiledRoute)
            and self.route_template == other.route_template
            and self.major_param_hash == other.major_param_hash
            and self.compiled_path == other.compiled_path
            and self.hash_code == other.hash_code
        )

    def __repr__(self) -> str:
        this_type = type(self).__name__
        major_params = ", ".join(
            (
                f"method={self.method!r}",
                f"compiled_path={self.compiled_path!r}",
                f"major_params_hash={self.major_param_hash!r}",
            )
        )
        return f"{this_type}({major_params})"

    def __str__(self) -> str:
        return f"{self.method} {self.compiled_path}"


class RouteTemplate:
    """A template used to create compiled routes for specific parameters.

    These compiled routes are used to identify rate limit buckets. Compiled
    routes may have a single major parameter.

    Parameters
    ----------
    method : str
        The HTTP method
    path_template : str
        The template string for the path to use.
    """

    _MAJOR_PARAM_REGEX = re.compile(r"\{(.*?)\}")

    __slots__ = ("method", "path_template", "major_param", "hash_code")

    method: str
    """The HTTP method."""

    path_template: typing.Final[str]
    """The template string used for the path."""

    major_param: typing.Final[typing.Optional[str]]
    """The optional major parameter name."""

    hash_code: typing.Final[int]
    """The hash code."""

    def __init__(self, method: str, path_template: str) -> None:
        self.method = method
        self.path_template = path_template

        if match := self._MAJOR_PARAM_REGEX.search(path_template):
            self.major_param = match.group(1)
        else:
            self.major_param = None

        self.hash_code = hash((self.method, self.path_template))

    def compile(self, **kwargs: typing.Any) -> CompiledRoute:
        """Generate a formatted `CompiledRoute` for this route template.

        This takes into account any URL parameters that have been passed.

        Parameters
        ----------
        **kwargs : typing.Any
            Any parameters to interpolate into the route path.

        Returns
        -------
        CompiledRoute
            The compiled route.
        """
        return CompiledRoute(
            self,
            self.path_template.format_map(kwargs),
            str(kwargs[self.major_param]) if self.major_param is not None else "-",
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}(path_template={self.path_template!r}, major_param={self.major_param!r})"

    def __str__(self) -> str:
        return self.path_template

    def __hash__(self) -> int:
        return self.hash_code

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, RouteTemplate)
            and self.method == other.method
            and self.major_param == other.major_param
            and self.path_template == other.path_template
            and self.hash_code == other.hash_code
        )


GET = "GET"
PATCH = "PATCH"
DELETE = "DELETE"
PUT = "PUT"
POST = "POST"

# Channels
GET_CHANNEL = RouteTemplate(GET, "/channels/{channel_id}")
PATCH_CHANNEL = RouteTemplate(PATCH, "/channels/{channel_id}")
DELETE_CHANNEL = RouteTemplate(DELETE, "/channels/{channel_id}")

GET_CHANNEL_INVITES = RouteTemplate(GET, "/channels/{channel_id}/invites")
POST_CHANNEL_INVITES = RouteTemplate(POST, "/channels/{channel_id}/invites")

GET_CHANNEL_MESSAGE = RouteTemplate(GET, "/channels/{channel_id}/messages/{message_id}")
PATCH_CHANNEL_MESSAGE = RouteTemplate(PATCH, "/channels/{channel_id}/messages/{message_id}")
DELETE_CHANNEL_MESSAGE = RouteTemplate(DELETE, "/channels/{channel_id}/messages/{message_id}")

GET_CHANNEL_MESSAGES = RouteTemplate(GET, "/channels/{channel_id}/messages")
POST_CHANNEL_MESSAGES = RouteTemplate(POST, "/channels/{channel_id}/messages")

POST_DELETE_CHANNEL_MESSAGES_BULK = RouteTemplate(POST, "/channels/{channel_id}/messages")

PATCH_CHANNEL_PERMISSIONS = RouteTemplate(PATCH, "/channels/{channel_id}/permissions/{overwrite_id}")
DELETE_CHANNEL_PERMISSIONS = RouteTemplate(DELETE, "/channels/{channel_id}/permissions/{overwrite_id}")

DELETE_CHANNEL_PIN = RouteTemplate(DELETE, "/channels/{channel_id}/pins/{message_id}")

GET_CHANNEL_PINS = RouteTemplate(GET, "/channels/{channel_id}/pins")
PUT_CHANNEL_PINS = RouteTemplate(PUT, "/channels/{channel_id}/pins")

POST_CHANNEL_TYPING = RouteTemplate(POST, "/channels/{channel_id}/typing")

POST_CHANNEL_WEBHOOKS = RouteTemplate(POST, "/channels/{channel_id}/webhooks")
GET_CHANNEL_WEBHOOKS = RouteTemplate(GET, "/channels/{channel_id}/webhooks")

# Reactions
DELETE_ALL_REACTIONS = RouteTemplate(DELETE, "/channels/{channel_id}/messages/{message_id}/reactions")

DELETE_REACTION_EMOJI = RouteTemplate(DELETE, "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}")
DELETE_REACTION_USER = RouteTemplate(DELETE, "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{used_id}")
GET_REACTIONS = RouteTemplate(GET, "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}")

# Guilds
GET_GUILD = RouteTemplate(GET, "/guilds/{guild_id}")
PATCH_GUILD = RouteTemplate(PATCH, "/guilds/{guild_id}")
DELETE_GUILD = RouteTemplate(DELETE, "/guilds/{guild_id}")

POST_GUILDS = RouteTemplate(POST, "/guilds")

GET_GUILD_AUDIT_LOGS = RouteTemplate(GET, "/guilds/{guild_id}/audit-logs")

GET_GUILD_BAN = RouteTemplate(GET, "/guilds/{guild_id}/bans/{user_id}")
PUT_GUILD_BAN = RouteTemplate(PUT, "/guilds/{guild_id}/bans/{user_id}")
DELETE_GUILD_BAN = RouteTemplate(DELETE, "/guilds/{guild_id}/bans/{user_id}")

GET_GUILD_BANS = RouteTemplate(GET, "/guilds/{guild_id}/bans")

GET_GUILD_CHANNELS = RouteTemplate(GET, "/guilds/{guild_id}/channels")
POST_GUILD_CHANNELS = RouteTemplate(POST, "/guilds/{guild_id}/channels")
PUT_GUILD_CHANNELS = RouteTemplate(PUT, "/guilds/{guild_id}/channels")
PATCH_GUILD_CHANNELS = RouteTemplate(PATCH, "/guilds/{guild_id}/channels")

GET_GUILD_EMBED = RouteTemplate(GET, "/guilds/{guild_id}/embed")
PATCH_GUILD_EMBED = RouteTemplate(PATCH, "/guilds/{guild_id}/embed")

GET_GUILD_EMOJI = RouteTemplate(GET, "/guilds/{guild_id}/emojis/{emoji_id}")
PATCH_GUILD_EMOJI = RouteTemplate(PATCH, "/guilds/{guild_id}/emojis/{emoji_id}")
DELETE_GUILD_EMOJI = RouteTemplate(DELETE, "/guilds/{guild_id}/emojis/{emoji_id}")

GET_GUILD_EMOJIS = RouteTemplate(GET, "/guilds/{guild_id}/emojis")
POST_GUILD_EMOJIS = RouteTemplate(POST, "/guilds/{guild_id}/emojis")

PATCH_GUILD_INTEGRATION = RouteTemplate(PATCH, "/guilds/{guild_id}/integrations/{integration_id}")
DELETE_GUILD_INTEGRATION = RouteTemplate(DELETE, "/guilds/{guild_id}/integrations/{integration_id}")

GET_GUILD_INTEGRATIONS = RouteTemplate(GET, "/guilds/{guild_id}/integrations")

POST_GUILD_INTEGRATION_SYNC = RouteTemplate(POST, "/guilds/{guild_id}/integrations/{integration_id}")

GET_GUILD_INVITES = RouteTemplate(GET, "/guilds/{guild_id}/invites")

GET_GUILD_MEMBERS = RouteTemplate(GET, "/guilds/{guild_id}/members")

GET_GUILD_MEMBER = RouteTemplate(GET, "/guilds/{guild_id}/members/{user_id}")
PATCH_GUILD_MEMBER = RouteTemplate(PATCH, "/guilds/{guild_id}/members/{user_id}")
DELETE_GUILD_MEMBER = RouteTemplate(DELETE, "/guilds/{guild_id}/members/{user_id}")

PUT_GUILD_MEMBER_ROLE = RouteTemplate(PUT, "/guilds/{guild_id}/members/{user_id}/roles/{role_id}")
DELETE_GUILD_MEMBER_ROLE = RouteTemplate(DELETE, "/guilds/{guild_id}/members/{user_id}/roles/{role_id}")

GET_GUILD_PREVIEW = RouteTemplate(GET, "/guilds/{guild_id}/preview")

GET_GUILD_PRUNE = RouteTemplate(GET, "/guilds/{guild_id}/prune")
POST_GUILD_PRUNE = RouteTemplate(POST, "/guilds/{guild_id}/prune")

PATCH_GUILD_ROLE = RouteTemplate(PATCH, "/guilds/{guild_id}/roles/{role_id}")
DELETE_GUILD_ROLE = RouteTemplate(DELETE, "/guilds/{guild_id}/roles/{role_id}")

GET_GUILD_ROLES = RouteTemplate(GET, "/guilds/{guild_id}/roles")
POST_GUILD_ROLES = RouteTemplate(POST, "/guilds/{guild_id}/roles")
PATCH_GUILD_ROLES = RouteTemplate(PATCH, "/guilds/{guild_id}/roles")

GET_GUILD_VANITY_URL = RouteTemplate(GET, "/guilds/{guild_id}/vanity-url")

GET_GUILD_VOICE_REGIONS = RouteTemplate(GET, "/guilds/{guild_id}/regions")

GET_GUILD_WIDGET_IMAGE = RouteTemplate(GET, "/guilds/{guild_id}/widget.png")

GET_GUILD_WEBHOOKS = RouteTemplate(GET, "/guilds/{guild_id}/webhooks")

# Invites
GET_INVITE = RouteTemplate(GET, "/invites/{invite_code}")
DELETE_INVITE = RouteTemplate(DELETE, "/invites/{invite_code}")

# Users
GET_USER = RouteTemplate(GET, "/users/{user_id}")

# @me
DELETE_MY_GUILD = RouteTemplate(DELETE, "/users/@me/guilds/{guild_id}")

GET_MY_CONNECTIONS = RouteTemplate(GET, "/users/@me/connections")  # OAuth2 only

POST_MY_CHANNELS = RouteTemplate(POST, "/users/@me/channels")

GET_MY_GUILDS = RouteTemplate(GET, "/users/@me/guilds")

PATCH_MY_GUILD_NICKNAME = RouteTemplate(PATCH, "/guilds/{guild_id}/members/@me/nick")

GET_MY_USER = RouteTemplate(GET, "/users/@me")
PATCH_MY_USER = RouteTemplate(PATCH, "/users/@me")

PUT_MY_REACTION = RouteTemplate(PUT, "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me")
DELETE_MY_REACTION = RouteTemplate(DELETE, "/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me")

# Voice
GET_VOICE_REGIONS = RouteTemplate(GET, "/voice/regions")

# Webhooks
GET_WEBHOOK = RouteTemplate(GET, "/webhooks/{webhook_id}")
PATCH_WEBHOOK = RouteTemplate(PATCH, "/webhooks/{webhook_id}")
POST_WEBHOOK = RouteTemplate(POST, "/webhooks/{webhook_id}")
DELETE_WEBHOOK = RouteTemplate(DELETE, "/webhooks/{webhook_id}")

GET_WEBHOOK_WITH_TOKEN = RouteTemplate(GET, "/webhooks/{webhook_id}/{webhook_token}")
PATCH_WEBHOOK_WITH_TOKEN = RouteTemplate(PATCH, "/webhooks/{webhook_id}/{webhook_token}")
DELETE_WEBHOOK_WITH_TOKEN = RouteTemplate(DELETE, "/webhooks/{webhook_id}/{webhook_token}")
POST_WEBHOOK_WITH_TOKEN = RouteTemplate(POST, "/webhooks/{webhook_id}/{webhook_token}")

POST_WEBHOOK_WITH_TOKEN_GITHUB = RouteTemplate(POST, "/webhooks/{webhook_id}/{webhook_token}/github")
POST_WEBHOOK_WITH_TOKEN_SLACK = RouteTemplate(POST, "/webhooks/{webhook_id}/{webhook_token}/slack")

# OAuth2 API
GET_MY_APPLICATION = RouteTemplate(GET, "/oauth2/applications/@me")

# Gateway
GET_GATEWAY = RouteTemplate(GET, "/gateway")
GET_GATEWAY_BOT = RouteTemplate(GET, "/gateway/bot")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities related to discord's Oauth2 flow."""

from __future__ import annotations

__all__ = [
    "Application",
    "ApplicationOwner",
    "ConnectionVisibility",
    "OAuth2Scope",
    "OwnConnection",
    "OwnGuild",
    "Team",
    "TeamMember",
    "TeamMembershipState",
]

import typing

import attr

from hikari import bases
from hikari import guilds
from hikari import permissions
from hikari import users
from hikari.internal import marshaller
from hikari.internal import more_enums
from hikari.internal import urls

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@more_enums.must_be_unique
class OAuth2Scope(str, more_enums.Enum):
    """OAuth2 Scopes that Discord allows.

    These are categories of permissions for applications using the OAuth2 API
    directly. Most users will only ever need the `BOT` scope when developing
    bots.
    """

    ACTIVITIES_READ = "activities.read"
    """Enable the app to fetch a user's "Now Playing/Recently Played" list.

    !!! note
        You must be whitelisted to use this scope.
    """

    ACTIVITIES_WRITE = "activities.write"
    """Enable the app to update a user's activity.

    !!! note
        You must be whitelisted to use this scope.

    !!! note
        This is not required to use the GameSDK activity manager.
    """

    APPLICATIONS_BUILDS_READ = "applications.builds.read"
    """Enable the app to read build data for a user's applications.

    !!! note
        You must be whitelisted to use this scope.
    """

    APPLICATIONS_BUILDS_UPLOAD = "applications.builds.upload"
    """Enable the app to upload/update builds for a user's applications.

    !!! note
        You must be whitelisted to use this scope.
    """

    APPLICATIONS_ENTITLEMENTS = "applications.entitlements"
    """Enable the app to read entitlements for a user's applications."""

    APPLICATIONS_STORE_UPDATE = "applications.store.update"
    """Enable the app to read and update store data for the user's applications.

    This includes store listings, achievements, SKU's, etc.

    !!! note
        The store API is deprecated and may be removed in the future.
    """

    BOT = "bot"
    """Used to add OAuth2 bots to a guild.

    !!! note
        This requires you to have set up a bot account for your application.
    """

    CONNECTIONS = "connections"
    """Enable the app to view third-party linked accounts such as Twitch."""

    EMAIL = "email"
    """Enable the app to view the user's email and application info."""

    GROUP_DM_JOIN = "gdm.join"
    """Enable the application to join users into a group DM."""

    GUILDS = "guilds"
    """Enable the app to view the guilds the user is in."""

    GUILDS_JOIN = "guilds.join"
    """Enable the app to add the user to a specific guild.

    !!! note
        This requires you to have set up a bot account for your application.
    """

    IDENTIFY = "identify"
    """Enable the app to view info about itself.

    !!! note
        This does not include email address info. Use the `EMAIL` scope instead
        to retrieve this information.
    """

    RELATIONSHIPS_READ = "relationships.read"
    """Enable the app to view a user's friend list.

    !!! note
        You must be whitelisted to use this scope.
    """

    RPC = "rpc"
    """Enable the RPC app to control the local user's Discord client.

    !!! note
        You must be whitelisted to use this scope.
    """

    RPC_API = "rpc.api"
    """Enable the RPC app to access the RPC API as the local user.

    !!! note
        You must be whitelisted to use this scope.
    """

    RPC_MESSAGES_READ = "messages.read"
    """Enable the RPC app to read messages from all channels the user is in."""

    RPC_NOTIFICATIONS_READ = "rpc.notifications.read"
    """Enable the RPC app to read  from all channels the user is in.

    !!! note
        You must be whitelisted to use this scope.
    """

    WEBHOOK_INCOMING = "webhook.incoming"
    """Used to generate a webhook that is returned in the OAuth2 token response.

    This is used during authorization code grants.
    """


@more_enums.must_be_unique
class ConnectionVisibility(int, more_enums.Enum):
    """Describes who can see a connection with a third party account."""

    NONE = 0
    """Only you can see the connection."""

    EVERYONE = 1
    """Everyone can see the connection."""


def _deserialize_integrations(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Sequence[guilds.GuildIntegration]:
    return [guilds.PartialGuildIntegration.deserialize(integration, **kwargs) for integration in payload]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class OwnConnection(bases.Entity, marshaller.Deserializable):
    """Represents a user's connection with a third party account.

    Returned by the `GET Current User Connections` endpoint.
    """

    id: str = marshaller.attrib(deserializer=str)
    """The string ID of the third party connected account.

    !!! warning
        Seeing as this is a third party ID, it will not be a snowflake.
    """

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The username of the connected account."""

    type: str = marshaller.attrib(deserializer=str, repr=True)
    """The type of service this connection is for."""

    is_revoked: bool = marshaller.attrib(raw_name="revoked", deserializer=bool, if_undefined=False, default=False)
    """Whether the connection has been revoked."""

    integrations: typing.Sequence[guilds.PartialGuildIntegration] = marshaller.attrib(
        deserializer=_deserialize_integrations, if_undefined=list, factory=list, inherit_kwargs=True,
    )
    """A sequence of the partial guild integration objects this connection has."""

    is_verified: bool = marshaller.attrib(raw_name="verified", deserializer=bool)
    """Whether the connection has been verified."""

    is_friend_syncing: bool = marshaller.attrib(raw_name="friend_sync", deserializer=bool)
    """Whether friends should be added based on this connection."""

    is_showing_activity: bool = marshaller.attrib(raw_name="show_activity", deserializer=bool)
    """Whether this connection's activities are shown in the user's presence."""

    visibility: ConnectionVisibility = marshaller.attrib(deserializer=ConnectionVisibility, repr=True)
    """The visibility of the connection."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class OwnGuild(guilds.PartialGuild):
    """Represents a user bound partial guild object."""

    is_owner: bool = marshaller.attrib(raw_name="owner", deserializer=bool, repr=True)
    """Whether the current user owns this guild."""

    my_permissions: permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=permissions.Permission
    )
    """The guild level permissions that apply to the current user or bot."""


@more_enums.must_be_unique
class TeamMembershipState(int, more_enums.Enum):
    """Represents the state of a user's team membership."""

    INVITED = 1
    """Denotes the user has been invited to the team but has yet to accept."""

    ACCEPTED = 2
    """Denotes the user has accepted the invite and is now a member."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class TeamMember(bases.Entity, marshaller.Deserializable):
    """Represents a member of a Team."""

    membership_state: TeamMembershipState = marshaller.attrib(deserializer=TeamMembershipState)
    """The state of this user's membership."""

    permissions: typing.Set[str] = marshaller.attrib(deserializer=set)
    """This member's permissions within a team.

    Will always be `["*"]` until Discord starts using this.
    """

    team_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of the team this member belongs to."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The user object of this team member."""


def _deserialize_members(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, TeamMember]:
    return {bases.Snowflake(member["user"]["id"]): TeamMember.deserialize(member, **kwargs) for member in payload}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Team(bases.Unique, marshaller.Deserializable):
    """Represents a development team, along with all its members."""

    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str)
    """The hash of this team's icon, if set."""

    members: typing.Mapping[bases.Snowflake, TeamMember] = marshaller.attrib(
        deserializer=_deserialize_members, inherit_kwargs=True
    )
    """The member's that belong to this team."""

    owner_user_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of this team's owner."""

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL of this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this team if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`. Can be any power
            of two between `16` and `4096` inclusive.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("team-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ApplicationOwner(users.User):
    """Represents the user who owns an application, may be a team user."""

    flags: int = marshaller.attrib(deserializer=users.UserFlag, repr=True)
    """This user's flags."""

    @property
    def is_team_user(self) -> bool:
        """If this user is a Team user (the owner of an application that's owned by a team)."""
        return bool((self.flags >> 10) & 1)


def _deserialize_verify_key(payload: str) -> bytes:
    return bytes(payload, "utf-8")


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Application(bases.Unique, marshaller.Deserializable):
    """Represents the information of an Oauth2 Application."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The name of this application."""

    description: str = marshaller.attrib(deserializer=str)
    """The description of this application, will be an empty string if unset."""

    is_bot_public: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_public", deserializer=bool, if_undefined=None, default=None, repr=True
    )
    """Whether the bot associated with this application is public.

    Will be `None` if this application doesn't have an associated bot.
    """

    is_bot_code_grant_required: typing.Optional[bool] = marshaller.attrib(
        raw_name="bot_require_code_grant", deserializer=bool, if_undefined=None, default=None
    )
    """Whether this application's bot is requiring code grant for invites.

    Will be `None` if this application doesn't have a bot.
    """

    owner: typing.Optional[ApplicationOwner] = marshaller.attrib(
        deserializer=ApplicationOwner.deserialize, if_undefined=None, default=None, inherit_kwargs=True, repr=True
    )
    """The object of this application's owner.

    This should always be `None` in application objects retrieved outside
    Discord's oauth2 flow.
    """

    rpc_origins: typing.Optional[typing.Set[str]] = marshaller.attrib(deserializer=set, if_undefined=None, default=None)
    """A collection of this application's rpc origin URLs, if rpc is enabled."""

    summary: str = marshaller.attrib(deserializer=str)
    """This summary for this application's primary SKU if it's sold on Discord.

    Will be an empty string if unset.
    """

    verify_key: typing.Optional[bytes] = marshaller.attrib(
        deserializer=_deserialize_verify_key, if_undefined=None, default=None
    )
    """The base64 encoded key used for the GameSDK's `GetTicket`."""

    icon_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="icon", deserializer=str, if_undefined=None, default=None
    )
    """The hash of this application's icon, if set."""

    team: typing.Optional[Team] = marshaller.attrib(
        deserializer=Team.deserialize, if_undefined=None, if_none=None, default=None, inherit_kwargs=True
    )
    """This application's team if it belongs to one."""

    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the guild this application is linked to if sold on Discord."""

    primary_sku_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the primary "Game SKU" of a game that's sold on Discord."""

    slug: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The URL slug that links to this application's store page.

    Only applicable to applications sold on Discord.
    """

    cover_image_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="cover_image", deserializer=str, if_undefined=None, default=None
    )
    """The hash of this application's cover image on it's store, if set."""

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL for this team's icon, if set."""
        return self.format_icon_url()

    def format_icon_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the icon URL for this application if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            return urls.generate_cdn_url("app-icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def cover_image_url(self) -> typing.Optional[str]:
        """URL for this icon's store cover image, if set."""
        return self.format_cover_image_url()

    def format_cover_image_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this application's store page's cover image is set and applicable.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.cover_image_hash:
            return urls.generate_cdn_url("app-assets", str(self.id), self.cover_image_hash, fmt=fmt, size=size)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling requests to `@me` endpoints."""

from __future__ import annotations

__all__ = ["RESTCurrentUserComponent"]

import abc
import datetime
import typing

from hikari import applications
from hikari import bases
from hikari import channels as _channels
from hikari import guilds
from hikari import pagination
from hikari import users
from hikari.clients.rest import base

if typing.TYPE_CHECKING:
    from hikari import files


class _GuildPaginator(pagination.BufferedPaginatedResults[guilds.Guild]):
    __slots__ = ("_session", "_components", "_newest_first", "_first_id")

    def __init__(self, newest_first, first_item, components, session):
        super().__init__()
        self._newest_first = newest_first
        self._first_id = self._prepare_first_id(
            first_item, bases.Snowflake.max() if newest_first else bases.Snowflake.min(),
        )
        self._components = components
        self._session = session

    async def _next_chunk(self):
        kwargs = {"before" if self._newest_first else "after": self._first_id}

        chunk = await self._session.get_current_user_guilds(**kwargs)

        if not chunk:
            return None

        self._first_id = chunk[-1]["id"]

        return (applications.OwnGuild.deserialize(g, components=self._components) for g in chunk)


class RESTCurrentUserComponent(base.BaseRESTComponent, abc.ABC):  # pylint: disable=abstract-method
    """The REST client component for handling requests to `@me` endpoints."""

    async def fetch_me(self) -> users.MyUser:
        """Get the current user that of the token given to the client.

        Returns
        -------
        hikari.users.MyUser
            The current user object.
        """
        payload = await self._session.get_current_user()
        return users.MyUser.deserialize(payload, components=self._components)

    async def update_me(self, *, username: str = ..., avatar: typing.Optional[files.BaseStream] = ...) -> users.MyUser:
        """Edit the current user.

        Parameters
        ----------
        username : str
            If specified, the new username string.
        avatar : hikari.files.BaseStream, optional
            If specified, the new avatar image data.
            If it is None, the avatar is removed.

        !!! warning
            Verified bots will not be able to change their username on this
            endpoint, and should contact Discord support instead to change
            this value.

        Returns
        -------
        hikari.users.MyUser
            The updated user object.

        Raises
        ------
        hikari.errors.BadRequest
            If you pass username longer than the limit (`2-32`) or an invalid image.
        """
        payload = await self._session.modify_current_user(
            username=username, avatar=await avatar.read() if avatar is not ... else ...,
        )
        return users.MyUser.deserialize(payload, components=self._components)

    async def fetch_my_connections(self) -> typing.Sequence[applications.OwnConnection]:
        """
        Get the current user's connections.

        !!! note
            This endpoint can be used with both `Bearer` and `Bot` tokens but
            will usually return an empty list for bots (with there being some
            exceptions to this, like user accounts that have been converted to
            bots).

        Returns
        -------
        typing.Sequence[hikari.applications.OwnConnection]
            A list of connection objects.
        """
        payload = await self._session.get_current_user_connections()
        return [
            applications.OwnConnection.deserialize(connection, components=self._components) for connection in payload
        ]

    def fetch_my_guilds(
        self,
        *,
        newest_first: bool = False,
        start_at: typing.Optional[typing.Union[datetime.datetime, bases.Unique, bases.Snowflake, int]] = None,
    ) -> pagination.PaginatedResults[applications.OwnGuild]:
        """Get an async iterable of the guilds the current user is in.

        Parameters
        ----------
        newest_first : bool
            If specified and `True`, the guilds are returned in the order of
            newest to oldest. The default is to return oldest guilds first.
        start_at : datetime.datetime OR hikari.bases.UniqueEntity OR hikari.bases.Snowflake or int, optional
            The optional first item to start at, if you want to limit your
            results. This will be interpreted as the date of creation for a
            guild. If unspecified, the newest or older possible snowflake is
            used, for `newest_first` being `True` and `False` respectively.

        Returns
        -------
        hikari.pagination.PaginatedResults[hikari.applications.OwnGuild]
            An async iterable of partial guild objects.

        Raises
        ------
        hikari.errors.NotFound
            If the guild is not found.
        """
        return _GuildPaginator(
            newest_first=newest_first, first_item=start_at, components=self._components, session=self._session
        )

    async def leave_guild(self, guild: typing.Union[bases.Snowflake, int, str, guilds.Guild]) -> None:
        """Make the current user leave a given guild.

        Parameters
        ----------
        guild : typing.Union[hikari.guilds.Guild, hikari.bases.Snowflake, int]
            The object or ID of the guild to leave.

        Raises
        ------
        hikari.errors.NotFound
            If the guild is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        await self._session.leave_guild(guild_id=str(guild.id if isinstance(guild, bases.Unique) else int(guild)))

    async def create_dm_channel(
        self, recipient: typing.Union[bases.Snowflake, int, str, users.User]
    ) -> _channels.DMChannel:
        """Create a new DM channel with a given user.

        Parameters
        ----------
        recipient : typing.Union[hikari.users.User, hikari.bases.Snowflake, int]
            The object or ID of the user to create the new DM channel with.

        Returns
        -------
        hikari.channels.DMChannel
            The newly created DM channel object.

        Raises
        ------
        hikari.errors.NotFound
            If the recipient is not found.
        hikari.errors.BadRequest
            If any invalid snowflake IDs are passed; a snowflake may be invalid
            due to it being outside of the range of a 64 bit integer.
        """
        payload = await self._session.create_dm(
            recipient_id=str(recipient.id if isinstance(recipient, bases.Unique) else int(recipient))
        )
        return _channels.DMChannel.deserialize(payload, components=self._components)
